# -*- coding: utf-8 -*-
################################################################################
# pcollections/_dict.py
# The persistent dict type for Python.
# By Noah C. Benson

from collections.abc import (
    Mapping,
    MutableMapping,
    Set,
    MappingView,
    KeysView,
    ItemsView,
    ValuesView,
    Sized
)

from phamt import (
    PHAMT,
    THAMT
)

from .abc import (
    PersistentMapping,
    TransientMapping
)


#===============================================================================
# pdict
# The persistent dict type.

class PDictView(MappingView):
    __slots__ = ('_mapping')
    def __new__(cls, d):
        if not isinstance(d, pdict):
            raise ValueError(f"can only make {cls} object from pdict")
        obj = super(PDictView, cls).__new__(cls)
        object.__setattr__(obj, '_mapping', d)
        return obj
    def __iter__(self):
        return map(lambda arg: self._from_kv(arg[1]), self._mapping._els)
    def __reversed__(self):
        return reversed(list(self.__iter__()))
    def __len__(self):
        return len(self._mapping)
    # Abstract methods that must be overloaded by the concrete view classes
    # below.
    def _from_kv(self, kv):
        raise NotImplementedError()
    def __contains__(self, k):
        raise NotImplementedError()
class pdict_keys(KeysView, PDictView):
    __slots__ = ()
    def _from_kv(self, kv):
        return kv[0]
    def __contains__(self, k):
        return (k in self._mapping)
    def __reversed__(self):
        return reversed(list(self.__iter__()))
    def __repr__(self):
        s = ", ".join(map(repr, iter(self)))
        return f"pdict_keys({{{s}}})"
    def __str__(self):
        s = ", ".join(map(repr, iter(self)))
        return f"pdict_keys({{{s}}})"
class pdict_items(ItemsView, PDictView):
    __slots__ = ()
    def _from_kv(self, kv):
        return kv
    def __contains__(self, kv):
        if not isinstance(kv, tuple) or len(kv) != 2:
            return False
        (k0,v0) = self._from_kv(kv)
        d = Ellipsis if v0 is None else None
        v = self._mapping.get(k0, d)
        (k,v) = self._from_kv((k0,v))
        return v0 == v
class pdict_values(ValuesView, PDictView):
    __slots__ = ()
    def _from_kv(self, kv):
        return kv[1]
    def __contains__(self, v):
        for ent in self._mapping._els:
            v0 = self._from_kv(ent[1][0])
            if v0 == v:
                return True
        return False
class pdict(PersistentMapping):
    """A persistent dict type similar to `dict`.

    `pdict()` returns the empty `pdict`.

    `pdict(iterable)` returns a `pdict` containing the elements in `iterable`,
    which must be `(key, value)` tuples.

    `pdict(mapping)` returns a `pdict` containing the key-value pairs in the
    given mapping.
    """
    empty = None
    @classmethod
    def _new(cls, els, idx, top):
        new_pdict = super(pdict, cls).__new__(cls)
        object.__setattr__(new_pdict, '_els', els)
        object.__setattr__(new_pdict, '_idx', idx)
        object.__setattr__(new_pdict, '_top', top)
        object.__setattr__(new_pdict, '_hashcode', None)
        return new_pdict
    __slots__ = ("_els", "_idx", "_top", "_hashcode")
    def __new__(cls, *args, **kw):
        n = len(args)
        if n == 1:
            arg = args[0]
        elif n == 0:
            if len(kw) == 0:
                return cls.empty
            else:
                arg = kw
                kw = {}
        else:
            msg = f"{cls.__name__} expects at most 1 argument, got {n}"
            raise TypeError(msg)
        # If any keyword options were given, we route through tdict.
        if not kw:
            # If arg is a tdict and no keyword arguments have been given, this
            # is a special case.
            if isinstance(arg, Sized) and len(arg) == 0:
                return cls.empty
            elif isinstance(arg, tdict):
                return cls._new(
                    arg._els.persistent(),
                    arg._idx.persistent(),
                    arg._top)
            elif isinstance(arg, cls):
                # Also, if it's already the right type, we can just return it
                # as-is.
                return arg
            elif isinstance(arg, pdict):
                return cls._new(arg._els, arg._idx, arg._top)
        # For anything else, however, we just route this through tdict.
        t = tdict(arg, **kw)
        return cls._new(
            t._els.persistent(),
            t._idx.persistent(),
            t._top)
    def __hash__(self):
        if self._hashcode is None:
            h = PersistentMapping.__hash__(self)
            object.__setattr__(self, '_hashcode', h)
        return self._hashcode
    def __len__(self):
        return len(self._els)
    def __contains__(self, k):
        h = hash(k)
        ii = self._idx.get(h, None)
        while ii is not None:
            ((kk,vv),ii) = self._els[ii]
            if k == kk:
                return True
        return False
    def __iter__(self):
        return map(lambda arg: arg[1][0][0], self._els)
    def __reversed__(self):
        return reversed(self.keys())
    def __getitem__(self, key):
        h = hash(key)
        ii = self._idx.get(h, None)
        while ii is not None:
            (kv,ii) = self._els[ii]
            if key == kv[0]:
                return kv[1]
        raise KeyError(key)
    def get(self, key, default=None):
        h = hash(key)
        ii = self._idx.get(h, None)
        while ii is not None:
            (kv,ii) = self._els[ii]
            if key == kv[0]:
                return kv[1]
        return default
    def transient(self):
        """Returns a transient copy of the dict in constant time."""
        return tdict._new(THAMT(self._els), THAMT(self._idx), self._top, self)
    def set(self, key, val):
        """Returns a copy of the pdict that maps the given key to the given
        value."""
        # Get the hash and initial index (if there is one).
        h = hash(key)
        ii = self._idx.get(h, None)
        if ii is None:
            # The object's hash is not here yet, so we can append to els and
            # insert it into idx.
            new_els = self._els.assoc(self._top, ((key, val), None))
            new_idx = self._idx.assoc(h, self._top)
        else:
            # First make sure it's not already in the dict.
            while ii is not None:
                (kv,ii_next) = self._els[ii]
                (k,v) = kv
                if key == k:
                    # It is in the dict; either it's exactly in the dict or we
                    # replace it.
                    if val is v:
                        return self
                    else:
                        new_els = self._els.assoc(ii, ((key,val), ii_next))
                        return self._new(new_els, self._idx, self._top)
            # If we reach this point, we can add the object to the end of the
            # list.
            new_els = self._els.assoc(self._top, ((key,val), None))
            new_els = new_els.assoc(ii, (kv, self._top))
        return self._new(new_els, new_idx, self._top + 1)
    def drop(self, key):
        """Returns a copy of the pdict that does not include the given key.

        If the key is not in the dict, returns the dict unchanged.
        """
        # Get the hash and initial index (if there is one).
        h = hash(key)
        ii = self._idx.get(h, None)
        # First make sure it's not already in the set.
        ii_prev = None
        kv_prev = None
        while ii is not None:
            (kv,ii_next) = self._els[ii]
            (k,v) = kv
            if key == k:
                # We remove this entry!
                if ii_prev is None:
                    # We're removing from the front of the list.
                    if ii_next is None:
                        new_idx = self._idx.dissoc(h)
                    else:
                        new_idx = self._idx.assoc(h, ii_next)
                    new_els = self._els
                else:
                    # We're removing from the end or the middle.
                    new_idx = self._idx
                    new_els = self._els.assoc(ii_prev, (kv_prev, ii_next))
                new_els = new_els.dissoc(ii)
                return self._new(new_els, new_idx, self._top)
            ii_prev = ii
            kv_prev = kv
            ii = ii_next
        # If we reach this point, then obj isn't in the set, so we just return
        # self unchanged.
        return self
    def clear(self):
        """Returns the empty pdict."""
        return type(self).empty
    # We include reimplements for some of these because we can improve them in
    # some non-trivial way.
    def pop(self, key, *args):
        """Returns a tuple of the value mapped to the given key and a copy of
        the pdict with that key removed. 

        If the key is not found, the second argument is returned, if given,
        otherwise, a `KeyError` is raised.
        """
        nargs = len(args)
        if nargs > 1:
            raise TypeError(f"pop expected at most 2 arguments, got {nargs}")
        h = hash(key)
        ii = self._idx.get(h, None)
        ii_prev = None
        kv_prev = None
        while ii is not None:
            (kv,ii_next) = self._els[ii]
            (k,v) = kv
            if key == k:
                # We remove this entry!
                if ii_prev is None:
                    # We're removing from the front of the list.
                    if ii_next is None:
                        new_idx = self._idx.dissoc(h)
                    else:
                        new_idx = self._idx.assoc(h, ii_next)
                    new_els = self._els
                else:
                    # We're removing from the end or the middle.
                    new_idx = self._idx
                    new_els = self._els.assoc(ii_prev, (kv_prev, ii_next))
                new_els = new_els.dissoc(ii)
                return (v, self._new(new_els, new_idx, self._top))
            ii_prev = ii
            kv_prev = kv
            ii = ii_next
        # It's not here!
        if nargs == 0:
            raise KeyError(key)
        else:
            return (args[0], self)
    def keys(self):
        return pdict_keys(self)
    def items(self):
        return pdict_items(self)
    def values(self):
        return pdict_values(self)
# Make the empty pdict.
pdict.empty = pdict._new(PHAMT.empty, PHAMT.empty, 0)


#===============================================================================
# tdict
# The transient set type.

class tdict_view(Set):
    def _iter(self, arg):
        if self._tdict._version > self._version:
            raise RuntimeError(f"{type(self)} changed during iteration")
        else:
            return self._from_kv(arg[1])
    def __new__(cls, d):
        if not isinstance(d, tdict):
            raise ValueError("can only make tdict_keys object from tdict")
        sup = super(tdict_view,cls)
        obj = sup.__new__(cls)
        sup.__setattr__(obj, '_tdict', d)
        sup.__setattr__(obj, '_version', d._version)
        return obj
    def __iter__(self):
        return map(self._iter, self._tdict._els)
    def __reversed__(self):
        return reversed(list(self.__iter__()))
    def __len__(self):
        return len(self._tdict)
    # These are the abstract methods that need to be implemented in the actual
    # view classes below.
    def _from_kv(kv):
        raise NotImplementedError()
    def __contains__(self, arg):
        raise NotImplementedError()
class tdict_keys(KeysView, tdict_view):
    __slots__ = ('_tdict', '_version')
    def _from_kv(self, arg):
        return arg[0]
    def __contains__(self, k):
        return (k in self._tdict)
class tdict_items(ItemsView, tdict_view):
    __slots__ = ('_tdict', '_version')
    def _from_kv(self, arg):
        return arg
    def __contains__(self, kv):
        if not isinstance(kv, tuple) or len(kv) != 2:
            return False
        d = Ellipsis if kv[1] is None else None
        return self._tdict.get(kv[0], d) == kv[1]
class tdict_values(ValuesView, tdict_view):
    __slots__ = ('_tdict', '_version')
    def _from_kv(self, arg):
        return arg[1]
    def __contains__(self, v):
        for (kv,_) in self._els:
            if kv[1] == v:
                return True
        return False
class tdict(TransientMapping):
    """A transient dict type similar to `dict` for mutating persistent dicts.

    `tdict()` returns the empty `tdict`.

    `tdict(iterable)` returns a `tdict` containing the key-value pairs in
    `iterable`.

    `tdict(p)` efficiently returns a transient copy of the persistent dict `p`.

    The interfaces for `dict` and `tdict` are nearly identical, but unlike
    `dict`, a `tdict` can be created from a `pdict` in `O(1)` time, and a
    `pdict` can be created from a `tdict` in `O(log n)` time (and note that in
    practice this `O(log n)` has a very low constant). The `tdict` type is
    explicitly intended to make batch mutations to `pdict` objects more
    efficient by reducing the number of allocations required.
    """
    @classmethod
    def _new(cls, els, idx, top, orig=None):
        new_tdict = super(tdict, cls).__new__(cls)
        object.__setattr__(new_tdict, '_els', els)
        object.__setattr__(new_tdict, '_idx', idx)
        object.__setattr__(new_tdict, '_top', top)
        object.__setattr__(new_tdict, '_version', 0)
        object.__setattr__(new_tdict, '_orig', orig)
        return new_tdict
    @classmethod
    def empty(cls):
        """Returns an empty tdict."""
        return cls._new(THAMT(PHAMT.empty), THAMT(PHAMT.empty), 0)
    __slots__ = ("_els", "_idx", "_top","_version")
    def __new__(cls, *args, **kw):
        n = len(args)
        if n == 1:
            arg = args[0]
        elif n == 0:
            if len(kw) == 0:
                return cls.empty()
            else:
                arg = kw
                kw = None
        else:
            raise TypeError(f"pdict expects at most 1 argument, got {n}")
        # If arg is a tdict or pdict, this is a special case.
        if isinstance(arg, tdict):
            obj = cls._new(THAMT(arg._els.persistent()),
                           THAMT(arg._idx.persistent()),
                           arg._top,
                           arg._orig)
        elif isinstance(arg, pdict):
            obj = cls._new(THAMT(arg._els), THAMT(arg._idx), arg._top, arg)
        else:
            obj = cls._new(THAMT(PHAMT.empty), THAMT(PHAMT.empty), 0)
            if isinstance(arg, Mapping):
                arg = arg.items()
            for (k,v) in arg:
                obj[k] = v
        # If keyword args were provided, merge them in.
        if kw:
            for (k,v) in kw.items():
                obj[k] = v
        return obj
    def __setattr__(self, k, v):
        raise TypeError("tdict attributes are immutable")
    def __setitem__(self, k, v):
        # Get the hash and initial index (if there is one).
        h = hash(k)
        ii = self._idx.get(h, None)
        # First make sure it's not already in the dict.
        ii_prev = None
        while ii is not None:
            (kv, ii_next) = self._els[ii]
            if kv[0] == k:
                if kv[1] is not v:
                    self._els[ii] = ((k,v), ii_next)
                    # The object has changed, so make sure we aren't tracking
                    # the original object still.
                    object.__setattr__(self, '_orig', None)
                # Note that this does not mandate a version update because it is
                # not changing the keys.
                return None
            ii_prev = ii
            kv_prev = kv
            ii = ii_next
        if ii_prev is None:
            # The object's hash is not here yet, so we can append to els and
            # insert it into idx.
            self._els[self._top] = ((k,v), None)
            self._idx[h] = self._top
        else:
            # If we reach this point, we can add the object to the end of the
            # list.
            self._els[self._top] = ((k,v), None)
            self._els[ii_prev] = (kv_prev, self._top)
        object.__setattr__(self, '_top', self._top + 1)
        object.__setattr__(self, '_version', self._version + 1)
        object.__setattr__(self, '_orig', None)
    def __len__(self):
        return len(self._els)
    def __contains__(self, k):
        h = hash(k)
        ii = self._idx.get(h, None)
        while ii is not None:
            ((kk,vv),ii) = self._els[ii]
            if k == kk:
                return True
        return False
    def __reversed__(self):
        return reversed(self.keys())
    def __getitem__(self, key):
        h = hash(key)
        ii = self._idx.get(h, None)
        while ii is not None:
            (kv,ii) = self._els[ii]
            if key == kv[0]:
                return kv[1]
        raise KeyError(key)
    def get(self, key, default=None):
        h = hash(key)
        ii = self._idx.get(h, None)
        while ii is not None:
            (kv,ii) = self._els[ii]
            if key == kv[0]:
                return kv[1]
        return default
    def __iter__(self):
        v0 = self._version
        def _iter_key(arg):
            if v0 < self._version:
                raise RuntimeError(f"{type(self)} changed during iteration")
            else:
                return arg[1][0][0]
        return map(_iter_key, self._els)
    def __delitem__(self, key):
        # Get the hash and initial index (if there is one).
        h = hash(key)
        ii = self._idx.get(h, None)
        # First make sure it's not already in the set.
        ii_prev = None
        kv_prev = None
        while ii is not None:
            (kv,ii_next) = self._els[ii]
            (k,v) = kv
            if key == k:
                # We remove this entry!
                if ii_prev is None:
                    # We're removing from the front of the list.
                    if ii_next is None:
                        del self._idx[h]
                    else:
                        self._idx[h] = ii_next
                else:
                    # We're removing from the end or the middle.
                    self._els[ii_prev] = (kv_prev, ii_next)
                del self._els[ii]
                object.__setattr__(self, '_version', self._version + 1)
                object.__setattr__(self, '_orig', None)
                return None
            ii_prev = ii
            kv_prev = kv
            ii = ii_next
        # If we reach this point, then obj isn't in the set, so we just return
        # self unchanged.
        raise KeyError(key)
    def clear(self):
        """Returns the empty pdict."""
        object.__setattr__(self, '_els', THAMT(PHAMT.empty))
        object.__setattr__(self, '_idx', THAMT(PHAMT.empty))
        object.__setattr__(self, '_top', 0)
        object.__setattr__(self, '_version', 0)
        return None
    def pop(self, key, *args):
        """Removes the given key from the tdict and returns the previously
        associated value.

        If the key is not found, the second argument is returned, if given,
        otherwise, a `KeyError` is raised.
        """
        nargs = len(args)
        if nargs > 1:
            raise TypeError(f"pop expected at most 2 arguments, got {nargs}")
        h = hash(key)
        ii = self._idx.get(h, None)
        ii_prev = None
        kv_prev = None
        while ii is not None:
            (kv,ii_next) = self._els[ii]
            (k,v) = kv
            if key == k:
                # We remove this entry!
                if ii_prev is None:
                    # We're removing from the front of the list.
                    if ii_next is None:
                        del self._idx[h]
                    else:
                        self._idx[h] = ii_next
                else:
                    # We're removing from the end or the middle.
                    self._els[ii_prev] = (kv_prev, ii_next)
                del self._els[ii]
                object.__setattr__(self, '_version', self._version + 1)
                object.__setattr__(self, '_orig', None)
                return v
            ii_prev = ii
            kv_prev = kv
            ii = ii_next
        # It's not here!
        if nargs == 0:
            raise KeyError(key)
        else:
            return args[0]
    def persistent(self):
        """Efficiently returns a persistent (pdict) copy of the tdict."""
        if len(self) == 0:
            return pdict.empty
        elif self._orig is not None:
            return self._orig
        else:
            return pdict._new(self._els.persistent(),
                              self._idx.persistent(),
                              self._top)
    def keys(self):
        return tdict_keys(self)
    def items(self):
        return tdict_items(self)
    def values(self):
        return tdict_values(self)



# -*- coding: utf-8 -*-
################################################################################
# pcollections/abc/_set.py
# The definitions of the abstract base classes for the pcollections set types.
# By Noah C. Benson

from collections.abc import (Set, MutableSet)

from ._core import (Persistent, Transient)
from ..util import (setcmp, seqstr)


#===============================================================================
# PersistentSet

class PersistentSet(Set, Persistent):
    """All the operations on a persistent set.

    Persistent sets are sets (i.e., objects that inherit from
    `collections.abc.Set`), but they differ from other sets in that they support
    efficient updating by means of efficiently producing copies of themselves
    that incorporate requested changes.

    The following abstract methods must be implemented; if these methods are
    inherited from a superclass of `PersistentSet`, that class is noted in
    parentheses.
     * `__iter__` (`Iterable`)
     * `__len__` (`Sized`)
     * `__contains__` (`Container`)
     * `transient()` (`Persistent`)
     * `add(index, object)`
     * `discard()`
     * `clear()`

    Additionally, `PersistentSequence` includes default implementations of the
    following methods, which may or may not be optimal for any particular
    base-class.
     * `__setattr__` (`object`; raises a `TypeError`)
     * `__setitem__` (`object`; raises a `TypeError`)
     * `__str__` (`object`)
     * `__repr__` (`object`)
     * `__eq__` (`object`)
     * `__ne__` (`object`)
     * `__hash__` (`Hashable`)
     * `__lt__` (`Set`)
     * `__le__` (`Set`)
     * `__gt__` (`Set`)
     * `__or__` (`Set`)
     * `__and__` (`Set`)
     * `__xor__` (`Set`)
     * `__sub__` (`Set`)
     * `__ror__` (`Set`)
     * `__rand__` (`Set`)
     * `__rxor__` (`Set`)
     * `__rsub__` (`Set`)
     * `isdisjoint` (`Set`)
     * `copy()` (`Persistent`)
     * `pop()`
     * `remove(value)`
     * `addall(values)`
     * `discardall(values)`
     * `removeall(values)`
     * `__reduce__` (for pickling)
    """
    # Methods which must be implemented in the children.
    def add(self, obj):
        """Returns a copy of the persistent set that includes the given
        object.
        """
        raise NotImplementedError()
    def discard(self, obj):
        """Returns a copy of the persistent set that does not include the given
        object.

        If the element is not a member, `discard` returns the original pset.
        """
        raise NotImplementedError()
    def clear(self):
        """Returns the empty pset."""
        raise NotImplementedError()
    # Methods which are probably fine for all child classes.
    def __str__(self):
        # We have a max length of 60 characters, not counting the delimiters.
        return f"{{|{seqstr(self, maxlen=60)}|}}"
    def __repr__(self):
        return f"{{|{seqstr(self)}|}}"
    def __eq__(self, other):
        if isinstance(other, Set):
            return setcmp(self, other) == 0
        else:
            return False
    def __ne__(self, other):
        if isinstance(other, Set):
            return setcmp(self, other) != 0
        else:
            return True
    def __hash__(self):
        return hash(frozenset(self)) + 1
    def __lt__(self, other):
        return setcmp(self, other) == -1
    def __le__(self, other):
        c = setcmp(self, other)
        return c == -1 or c == 0
    def __gt__(self, other):
        return setcmp(self, other) == 1
    def __ge__(self, other):
        c = setcmp(self, other)
        return c == 1 or c == 0
    def __and__(self, other):
        if not isinstance(other, Set):
            raise TypeError(f"unsupported operand type for &:"
                            f" '{type(self)}' and '{type(other)}'")
        t = self.transient()
        t &= other
        return type(self)(t)
    def __or__(self, other):
        if not isinstance(other, Set):
            raise TypeError(f"unsupported operand type for |:"
                            f" '{type(self)}' and '{type(other)}'")
        t = self.transient()
        t |= other
        return type(self)(t)
    def __sub__(self, other):
        if not isinstance(other, Set):
            raise TypeError(f"unsupported operand type for -:"
                            f" '{type(self)}' and '{type(other)}'")
        t = self.transient()
        t -= other
        return type(self)(t)
    def __xor__(self, other):
        if not isinstance(other, Set):
            raise TypeError(f"unsupported operand type for ^:"
                            f" '{type(self)}' and '{type(other)}'")
        t = self.transient()
        t ^= other
        return type(self)(t)
    def __rand__(self, other):
        if not isinstance(other, Set):
            raise TypeError(f"unsupported operand type for &:"
                            f" '{type(other)}' and '{type(set)}'")
        t = self.transient()
        t &= other
        return type(self)(t)
    def __ror__(self, other):
        if not isinstance(other, Set):
            raise TypeError(f"unsupported operand type for |:"
                            f" '{type(other)}' and '{type(set)}'")
        t = self.transient()
        t |= other
        return type(self)(t)
    def __rsub__(self, other):
        if not isinstance(other, Set):
            raise TypeError(f"unsupported operand type for -:"
                            f" '{type(other)}' and '{type(set)}'")
        t = self.clear().transient()
        t.addall(other)
        t -= self
        return type(self)(t)
    def __rxor__(self, other):
        if not isinstance(other, Set):
            raise TypeError(f"unsupported operand type for ^:"
                            f" '{type(other)}' and '{type(set)}'")
        t = self.transient()
        t ^= other
        return type(self)(t)
    def isdisjoint(self, other):
        """Returns `True` if two sets have a null intersection."""
        if not isinstance(other, Set):
            other = set(other)
        return setcmp(self, other) is None
    def issubset(self, other):
        """Report whether another set contains this set."""
        if not isinstance(other, (set, frozenset)):
            other = set(other)
        return other.issuperset(self)
    def issuperset(self, other):
        """Report whether this set contains another set."""
        if isinstance(other, (set, frozenset)):
            return other.issubset(self)
        else:
            return all(map(self.__contains__, iter(other)))
    def difference(self, *args):
        """Returns the difference of two or more sets as a new persistent set.

        (I.e., all elements that are in this set but not the others.)
        """
        t = self.transient()
        for arg in args:
            t.discardall(arg)
        if len(self) == len(t):
            return self
        return type(self)(t)
    def intersection(self, *args):
        """Return the intersection of two sets as a new persistent set.

        (I.e., all elements that are in both sets.)
        """
        t = self.transient()
        for arg in args:
            t &= arg
        if len(t) == len(self):
            return self
        return type(self)(t)
    def symmetric_difference(self, *args):
        """Return the symmetric difference of two sets as a new set.

        (I.e., all elements that are in exactly one of the sets.)
        """
        t = self.transient()
        for arg in args:
            t ^= arg
        if len(t) == len(self):
            return self
        return type(self)(t)
    def union(self, *args):
        """Returns the union of the persistent set and all arguments."""
        t = self.transient()
        for arg in args:
            t.addall(arg)
        if len(t) == len(self):
            return self
        else:
            return type(self)(t)
    def pop(self):
        """Returns a tuple of an arbitrary element from the persistent set and a
        copy of the set with that element removed.

        Raises `KeyError` if the set is empty.
        """
        if len(self) == 0:
            raise KeyError(f"pop from empty {type(self)}")
        el = next(iter(self))
        newset = self.discard(el)
        return (el, newset)
    def remove(self, element):
        """Removes an element from the persistent set; it must be a member.

        Raises a `KeyError` if the element is not a member.
        """
        s = self.discard(element)
        if len(s) == len(self):
            raise KeyError(element)
        return s
    def addall(self, iterable):
        """Returns a copy of the persistent set with the given elements
        included.
        """
        t = self.transient()
        t.addall(iterable)
        if len(t) == len(self):
            return self
        else:
            return type(self)(t)
    def discardall(self, iterable):
        """Returns a copy of the persistent set with the given elements
        discarded.

        If any element is not found, then it is ignored.
        """
        t = self.transient()
        t.discardall(iterable)
        if len(t) == len(self):
            return self
        else:
            return type(self)(t)
    def removeall(self, iterable):
        """Returns a copy of the persistent set with the given elements
        discarded.

        If any element is not found, then it is ignored.
        """
        t = self.transient()
        t.removeall(iterable)
        if len(t) == len(self):
            return self
        else:
            return type(self)(t)
    def __reduce__(self):
        return (self.__new__, (type(self), list(self),))


#===============================================================================
# TransientSet

class TransientSet(MutableSet, Transient):
    """All the operations on a transient set.

    Transient sets are mutable sets (i.e., objects that inherit from
    `collections.abc.MutableSet`), but they differ from other mutable sets in
    that they support efficient updating by means of efficiently producing
    copies of themselves that incorporate requested changes.

    The following abstract methods must be implemented; if these methods are
    inherited from a superclass of `TransientSet`, that class is noted in
    parentheses.
     * `__len__` (`Sized`)
     * `__contains__` (`Container`)
     * `__iter__` (`Iterable`)
     * `add(object)` (`MutableSet`)
     * `discard()`  (`MutableSet`)
     * `persistent()` (`Transient`)
     * `clear()`
     * `__getstate__` (for pickling)
     * `__setstate__` (for pickling)

    Additionally, `PersistentSequence` includes default implementations of the
    following methods, which may or may not be optimal for any particular
    base-class.
     * `__setattr__` (`object`; raises a `TypeError`)
     * `__setitem__` (`object`; raises a `TypeError`)
     * `__str__` (`object`)
     * `__repr__` (`object`)
     * `__eq__` (`object`)
     * `__ne__` (`object`)
     * `__hash__` (`Hashable`)
     * `__lt__` (`Set`)
     * `__le__` (`Set`)
     * `__gt__` (`Set`)
     * `__ge__` (`Set`)
     * `isdisjoint` (`Set`)
     * `__or__` (`Set`)
     * `__and__` (`Set`)
     * `__xor__` (`Set`)
     * `__sub__` (`Set`)
     * `__ror__` (`Set`)
     * `__rand__` (`Set`)
     * `__rxor__` (`Set`)
     * `__rsub__` (`Set`)
     * `__ior__` (`MutableSet`)
     * `__iand__` (`MutableSet`)
     * `__ixor__` (`MutableSet`)
     * `__isub__` (`MutableSet`)
     * `copy()` (`Transient`)
     * `pop()`
     * `remove(value)`
     * `addall(values)`
     * `discardall(values)`
     * `removeall(values)`
    """
    # Methods which must be implemented in the children.
    def add(self, obj):
        """Adds the given object to the transient set."""
        raise NotImplementedError()
    def discard(self, obj):
        """Discards the given object from the transient set.

        If the element is not a member, `discard` does nothing.
        """
        raise NotImplementedError()
    def clear(self):
        """Clears the transient set."""
        raise NotImplementedError()
    # Methods which are probably fine for all child classes.
    def __str__(self):
        # We have a max length of 60 characters, not counting the delimiters.
        return f"{{<{seqstr(self, maxlen=60)}>}}"
    def __repr__(self):
        #s = repr(dict(self))
        #return f"{{<{s[1:-1]}>}}"
        return f"{{<{seqstr(self)}>}}"
    def __eq__(self, other):
        return setcmp(self, other) == 0
    def __ne__(self, other):
        return setcmp(self, other) != 0
    def __hash__(self):
        return hash(frozenset(self)) + 1
    def __lt__(self, other):
        return setcmp(self, other) == -1
    def __le__(self, other):
        c = setcmp(self, other)
        return c == -1 or c == 0
    def __gt__(self, other):
        return setcmp(self, other) == 1
    def __ge__(self, other):
        c = setcmp(self, other)
        return c == 1 or c == 0
    def isdisjoint(self, other):
        """Returns `True` if two sets have a null intersection."""
        return setcmp(self, other) is None
    def issubset(self, other):
        """Report whether another set contains this set."""
        if not isinstance(other, (set, frozenset)):
            other = set(other)
        return other.issuperset(self)
    def issuperset(self, other):
        """Report whether this set contains another set."""
        if isinstance(other, (set, frozenset)):
            return other.issubset(self)
        else:
            return all(map(self.__contains__, iter(other)))
    def difference_update(self, *args):
        """Remove all elements of another set from this set."""
        for arg in args:
            self.discardall(arg)
    def difference(self, *args):
        """Returns the difference of two or more sets as a new transient set.

        (I.e., all elements that are in this set but not the others.)
        """
        t = self.copy()
        t.difference_update(*args)
        return t
    def intersection_update(self, *args):
        """Update a set with the intersection of itself and another."""
        for arg in args:
            self &= arg
    def intersection(self, *args):
        """Return the intersection of two sets as a new transient set.

        (I.e., all elements that are in both sets.)
        """
        t = self.copy()
        t.intersection_update(*args)
        return t
    def symmetric_difference_update(self, other):
        """Update a set with the symmetric difference of itself and another."""
        self ^= arg
    def symmetric_difference(self, other):
        """Return the symmetric difference of two sets as a new set.

        (I.e., all elements that are in exactly one of the sets.)
        """
        t = self.copy()
        t.symmetric_difference_update(other)
        return t
    def update(self, *args):
        for arg in args:
            self.addall(arg)
    def union(self, *args):
        """Returns the union of the transient set and all arguments."""
        t = self.copy()
        t.update(*args)
        return t
    def __and__(self, other):
        if not isinstance(other, Set):
            raise TypeError(f"unsupported operand type for &:"
                            f" '{type(self)}' and '{type(other)}'")
        t = self.copy()
        t &= other
        return t
    def __or__(self, other):
        if not isinstance(other, Set):
            raise TypeError(f"unsupported operand type for |:"
                            f" '{type(self)}' and '{type(other)}'")
        t = self.copy()
        t |= other
        return t
    def __sub__(self, other):
        if not isinstance(other, Set):
            raise TypeError(f"unsupported operand type for -:"
                            f" '{type(self)}' and '{type(other)}'")
        t = self.copy()
        t -= other
        return t
    def __xor__(self, other):
        if not isinstance(other, Set):
            raise TypeError(f"unsupported operand type for ^:"
                            f" '{type(self)}' and '{type(other)}'")
        t = self.copy()
        t ^= other
        return t
    def __rand__(self, other):
        return self.__and__(other)
    def __ror__(self, other):
        return self.__or__(other)
    def __rsub__(self, other):
        if not isinstance(other, Set):
            raise TypeError(f"unsupported operand type for -:"
                            f" '{type(other)}' and '{type(set)}'")
        t = self.copy()
        t.clear()
        t.addall(other)
        t -= self
        return t
    def __rxor__(self, other):
        return self.__xor__(other)
    def __iand__(self, other):
        if not isinstance(other, Set):
            raise TypeError(f"unsupported operand type for &=:"
                            f" '{type(self)}' and '{type(other)}'")
        rm = [el for el in self if el not in other]
        for el in rm:
            self.remove(el)
        return self
    def __ior__(self, other):
        if not isinstance(other, Set):
            raise TypeError(f"unsupported operand type for |=:"
                            f" '{type(self)}' and '{type(other)}'")
        self.addall(other)
        return self
    def __isub__(self, other):
        if not isinstance(other, Set):
            raise TypeError(f"unsupported operand type for -=:"
                            f" '{type(self)}' and '{type(other)}'")
        self.discardall(other)
        return self
    def __ixor__(self, other):
        if not isinstance(other, Set):
            raise TypeError(f"unsupported operand type for ^=:"
                            f" '{type(self)}' and '{type(other)}'")
        rm = []
        for el in self:
            if el in other:
                rm.append(el)
        for el in other:
            if el in self:
                self.remove(el)
            else:
                self.add(el)
        for el in rm:
            self.discard(el)
        return self
    def pop(self):
        """Removes an arbitrary element from the transient set and returns it.

        Raises `KeyError` if the set is empty.
        """
        if len(self) == 0:
            raise KeyError(f"pop from empty {type(self)}")
        el = next(iter(self))
        self.discard(el)
        return el
    def remove(self, element):
        """Removes an element from the persistent set; it must be a member.

        Raises a `KeyError` if the element is not a member.
        """
        n = len(self)
        self.discard(element)
        if n == len(self):
            raise KeyError(element)
    def addall(self, iterable):
        """Returns a copy of the persistent set with the given elements
        included.
        """
        for el in iterable:
            self.add(el)
    def discardall(self, iterable):
        """Returns a copy of the persistent set with the given elements
        discarded.

        If any element is not found, then it is ignored.
        """
        for el in iterable:
            self.discard(el)
    def removeall(self, iterable):
        """Returns a copy of the persistent set with the given elements
        discarded.

        If any element is not found, then a `KeyError` is raised.
        """
        for el in iterable:
            self.remove(el)
    def copy(self):
        """Returns a copy of the transient set."""
        return self.persistent().transient()
    def __reduce__(self):
        return (self.__new__, (type(self), list(self),))



# -*- coding: utf-8 -*-
################################################################################
# pcollections/test/__init__.py
# Declaration of tests for the pcollections package.
# By Noah C. Benson

from ._list import (TestPList, TestLList)
from ._set import TestPSet
from ._dict import (TestPDict, TestLDict)



# -*- coding: utf-8 -*-
################################################################################
# pcollections/__init__.py
# Initialization file for the pcollections library.
# By Noah C. Benson

"""Persistent and Transient Collections for Python
"""

from ._list import (plist, tlist)
from ._set  import (pset,  tset)
from ._dict import (pdict, tdict)
from ._lazy import (lazy, unlazy, holdlazy, llist, ldict)

# We don't include the abc types in the __all__; they are probably not as
# frequently used and don't really need to be here. One can always `import
# pcollections.abc` if they are needed.
# 
# from .abc import (
#     Persistent,         Transient,
#     PersistentSequence, TransientSequence,
#     PersistentSet,      TransientSet,
#     PersistentMapping,  TransientMapping)

__all__ = (
    "plist", "tlist",
    "pset",  "tset",
    "pdict", "tdict",
    "llist", "ldict",
    "lazy", "unlazy", "holdlazy")

__version__ = "0.3.1"




# -*- coding: utf-8 -*-
################################################################################
# pcollections/abc/_core.py
# The definitions of the abstract base classes for the persistent data types.
# By Noah C. Benson

from collections.abc import Hashable

class Persistent(Hashable):
    # Abstract methods.
    def transient(self):
        """Efficiently returns a transient copy of the persistent object."""
        raise NotImplementedError()
    # Methods that should throw errors in children.
    def __setattr__(self, k, v):
        raise TypeError(f"{type(self)} is immutable")
    def __setitem__(self, k, v):
        raise TypeError(f"{type(self)} is immutable")
    def __delitem__(self, k, v):
        raise TypeError(f"{type(self)} is immutable")
    # Implementation methods that are probably fine in all children.
    def copy(self):
        """Returns the persistent object (persistent objects needn't be
        copied)."""
        return self

class Transient:
    def persistent(self):
        """Efficiently returns a persistent copy of the transient object."""
        raise NotImplementedError()
    # Implementations that are probably fine for most children.
    def copy(self):
        """Returns a copy of the transient object."""
        return self.persistent().transient()
    



# -*- coding: utf-8 -*-
################################################################################
# pcollections/_list.py
# The persistent list type for Python.
# By Noah C. Benson

from itertools import (chain, islice)

from phamt import (PHAMT,THAMT)

from .abc import (PersistentSequence, TransientSequence)


#===============================================================================
# plist

class plist(PersistentSequence):
    """A persistent list type similar to `list`.

    `plist()` returns an empty `plist`.

    `plist(iterable)` returns a `plist` containing the elements in `iterable`.
    """
    empty = None
    __slots__ = ("_phamt", "_start", "_hashcode")
    def __new__(cls, *args, **kw):
        if len(kw) > 0:
            raise TypeError(f"{cls.__name__}() takes no keyword arguments")
        n = len(args)
        if n == 0:
            return cls.empty
        elif n != 1:
            msg = f"{cls.__name__} expects at most 1 argument, got {n}"
            raise TypeError(msg)
        arg = args[0]
        # If arg is a tlist, this is a special case.
        if isinstance(arg, tlist):
            if len(arg) == 0:
                return cls.empty
            else:
                return cls._new(arg._thamt.persistent(), arg._start)
        elif isinstance(arg, cls):
            return arg
        elif isinstance(arg, plist):
            if len(arg) == 0:
                return cls.empty
            else:
                return cls._new(arg._phamt, arg._start)        
        # We just want to build a PHAMT out of this arg of iterables.
        thamt = THAMT(PHAMT.empty)
        for (ii,val) in enumerate(iter(arg)):
            thamt[ii] = val
        phamt = thamt.persistent()
        # If it's empty, we can just return the empty plist.
        if len(phamt) == 0: return cls.empty
        # Otherwise, we make a new plist and give it this phamt.
        return cls._new(phamt, 0)
    @classmethod
    def _new(cls, phamt, start):
        new_plist = super(plist, cls).__new__(cls)
        object.__setattr__(new_plist, '_phamt', phamt)
        object.__setattr__(new_plist, '_start', start)
        object.__setattr__(new_plist, '_hashcode', None)
        return new_plist
    def set(self, index, obj):
        """Returns a copy of the list with the given index set to the given
        object."""
        start = self._start
        phamt = self._phamt
        n = len(phamt)
        if index < -n or index >= n:
            raise IndexError("plist index out of range")
        if index < 0:
            index += n
        index += start
        if phamt[index] is obj:
            return self
        new_phamt = phamt.assoc(index, obj)
        return self._new(new_phamt, start)
    def delete(self, index=-1):
        """"Returns a copy of the plist with the item at index removed (default
        index: last)."""
        phamt = self._phamt
        n = len(phamt)
        st = self._start
        if n == 0:
            raise IndexError("delete from empty plist")
        if index >= n or index < -n:
            raise IndexError("plist.delete index out of range")
        elif index < 0:
            index += n
        # Some cases don't require a THAMT:
        if index == 0:
            if n == 1:
                return plist.empty
            else:
                return plist._new(phamt.dissoc(st), st + 1)
        elif index == n - 1:
            return plist._new(phamt.dissoc(st + index), st)
        # Other cases require moving elements around a fair bit, so we use a
        # THAMT object.
        thamt = THAMT(phamt)
        if n - index <= index:
            for ii in range(index + st, n + st - 1):
                thamt[ii] = phamt[ii + 1]
            del thamt[n + st - 1]
        else:
            for ii in range(st, index + st):
                thamt[ii + 1] = phamt[ii]
            del thamt[st]
            st += 1
        return self._new(thamt.persistent(), st)
    def append(self, obj):
        """Returns a new list with object appended."""
        phamt = self._phamt
        n = len(phamt)
        new_phamt = phamt.assoc(self._start + n, obj)
        return self._new(new_phamt, self._start)
    def prepend(self, obj):
        """Returns a new list with object prepended."""
        phamt = self._phamt
        n = len(phamt)
        new_start = self._start - 1
        new_phamt = phamt.assoc(new_start, obj)
        return self._new(new_phamt, new_start)
    def insert(self, index, obj):
        """Returns a new plist with object inserted before index."""
        start = self._start
        phamt = self._phamt
        n = len(phamt)
        if   index == 0: return self.prepend(obj)
        elif index == n: return self.append(obj)
        elif index < -n: index = -n
        elif index > n:  index = n
        if   index < 0:  index += n
        thamt = THAMT(phamt)
        if n - index <= index:
            for ii in range(index + start, n + start):
                thamt[ii + 1] = phamt[ii]
            thamt[index + start] = obj
        else:
            for ii in range(start, index + start):
                thamt[ii - 1] = phamt[ii]
            start -= 1
            thamt[index + start] = obj
        phamt = thamt.persistent()
        return self._new(phamt, start)
    def clear(self):
        """Returns the empty plist."""
        return plist.empty
    def __iter__(self):
        phamt = self._phamt
        n = len(phamt)
        st = self._start
        if st >= 0 or n + st < 0:
            return map(lambda u:u[1], self._phamt)
        else:
            from itertools import chain, islice
            return chain((phamt[k] for k in range(st, 0)),
                         islice(map(lambda u:u[1], phamt), 0, n + st))
    def __len__(self):
        """Returns the length of the plist."""
        return len(self._phamt)
    def __getitem__(self, k):
        st = self._start
        phamt = self._phamt
        n = len(phamt)
        if isinstance(k, slice):
            (start,stop,step) = (k.start or 0, k.stop or n, k.step or 1)
            if   start <= -n: start = 0
            elif start < 0:   start += n
            elif start > n:   start = n
            if   stop <= -n: stop = 0
            elif stop < 0:   stop += n
            elif stop > n:   stop = n
            if st != 0:
                start += st
                stop += st
            thamt = THAMT(PHAMT.empty)
            for (ii,jj) in enumerate(range(start, stop, step)):
                thamt[ii] = phamt[jj]
            return self._new(thamt.persistent(), 0)
        elif k >= n or k < -n:
            raise IndexError("plist index out of range")
        elif k < 0:
            k += n
        return phamt[k + st]
    def transient(self):
        """Efficiently copes the plist into a tlist and returns the tlist."""
        return tlist._new(THAMT(self._phamt), self._start, self)
    # We redefine the hash function in order to use the _hashcode member.
    def __hash__(self):
        if self._hashcode is None:
            h = PersistentSequence.__hash__(self)
            object.__setattr__(self, '_hashcode', h)
        return self._hashcode
# Setup the plist.empty static member.
plist.empty = plist._new(PHAMT.empty, 0)


#===============================================================================
# tlist
# The transient list type.

class tlist(TransientSequence):
    """A transient list type, similar to `list`, for mutating persistent lists.

    `tlist()` returns an empty `tlist`.

    `tlist(iterable)` returns a `tlist` containing the elements in `iterable`.

    `tlist(p)` returns `tlist` equivalent to the `plist` p in constnat time.
    The returned `tlist` object can be efficiently mutated in-place then
    efficiently converted into a persistent list by calling the `persistent()`
    method.
    """
    @classmethod
    def empty(cls):
        "Returns an empty tlist."
        return cls._new(THAMT(PHAMT.empty), 0)
    __slots__ = ("_thamt", "_start", "_orig")
    def __new__(cls, *args, **kw):
        if len(kw) > 0:
            raise TypeError("tlist() takes no keyword arguments")
        n = len(args)
        if   n == 1: pass
        elif n == 0: return cls.empty()
        else: raise TypeError(f"tlist expects at most 1 argument, got {n}")
        arg = args[0]
        # If this is a plist, we know what to do with it.
        if isinstance(arg, plist):
            return cls._new(THAMT(arg._phamt), arg._start)
        # We just want to build a THAMT out of this arg of iterables.
        thamt = THAMT(PHAMT.empty)
        for (ii,val) in enumerate(iter(arg)):
            thamt[ii] = val
        # We make a new tlist and give it this phamt.
        return cls._new(thamt, 0)
    @classmethod
    def _new(cls, thamt, start, orig=None):
        new_tlist = super(tlist, cls).__new__(cls)
        object.__setattr__(new_tlist, '_thamt', thamt)
        object.__setattr__(new_tlist, '_start', start)
        object.__setattr__(new_tlist, '_orig', orig)
        return new_tlist
    def clear(self):
        """Clears all elements from the tlist."""
        object.__setattr__(self, '_thamt', THAMT(PHAMT.empty))
        object.__setattr__(self, '_start', 0)
        object.__setattr__(self, '_orig', None)
    def persistent(self):
        """Efficiently copies the tlist into a plist and returns the plist."""
        if len(self._thamt) == 0:
            return plist.empty
        elif self._orig is None:
            return plist._new(self._thamt.persistent(), self._start)
        else:
            return self._orig
    def __iter__(self):
        st = self._start
        th = self._thamt
        n = len(th)
        if st >= 0 or n + st < 0:
            return map(lambda u:u[1], iter(th))
        else:
            return chain((th[k] for k in range(st, 0)),
                         islice(map(lambda u:u[1], iter(th)), 0, n + st))
    def __len__(self):
        """Returns the length of the tlist."""
        return len(self._thamt)
    def __getitem__(self, k):
        st = self._start
        thamt = self._thamt
        n = len(thamt)
        if isinstance(k, slice):
            (start,stop,step) = (k.start or 0, k.stop or n, k.step or 1)
            if   start <= -n: start = 0
            elif start < 0:   start += n
            elif start > n:   start = n
            if   stop <= -n: stop = 0
            elif stop < 0:   stop += n
            elif stop > n:   stop = n
            if st != 0:
                start += st
                stop += st
            new_thamt = THAMT(PHAMT.empty)
            for (ii,jj) in enumerate(range(start, stop, step)):
                new_thamt[ii] = thamt[jj]
            return self._new(new_thamt, 0)
        elif k >= n or k < -n:
            raise IndexError("tlist index out of range")
        elif k < 0:
            k += n
        return thamt[k + st]
    def __setitem__(self, k, v):
        n = len(self._thamt)
        if k >= n or k < -n:
            raise IndexError(k)
        elif k < 0:
            k += n
        self._thamt[k + self._start] = v
        object.__setattr__(self, '_orig', None)
    def __delitem__(self, index=-1):
        """Remove and return item at index (default last).

        Raises IndexError if list is empty or index is out of range."""
        st = self._start
        th = self._thamt
        n = len(th)
        if index >= n or index < -n:
            raise IndexError(f"{type(self)} assignment index out of range")
        elif index < 0:
            index += n
        if n - index <= index:
            for ii in range(index + st, n + st - 1):
                th[ii] = th[ii + 1]
            del th[n + st - 1]
        else:
            for ii in range(index + st, st, -1):
                th[ii] = th[ii - 1]
            del th[st]
            self._start += 1
        object.__setattr__(self, '_orig', None)
    def append(self, obj):
        """Appends object to the end of the list."""
        thamt = self._thamt
        n = len(thamt)
        thamt[n + self._start] = obj
        object.__setattr__(self, '_orig', None)
    def prepend(self, obj):
        """Prepends object to the beginning of the tlist."""
        thamt = self._thamt
        n = len(thamt)
        self._start -= 1
        thamt[self._start] = obj
        object.__setattr__(self, '_orig', None)
    def insert(self, index, obj):
        """Inserts the given object before the given index."""
        st = self._start
        th = self._thamt
        n = len(th)
        if   index < -n: index = -n
        elif index > n:  index = n
        if   index < 0:  index += n
        if n - index <= index:
            for ii in range(n + st, index + st, -1):
                th[ii] = th[ii - 1]
            th[index + st] = obj
        else:
            for ii in range(st - 1, index + st - 1):
                th[ii] = th[ii + 1]
            self._start -= 1
            th[index + self._start] = obj
        object.__setattr__(self, '_orig', None)



# -*- coding: utf-8 -*-
################################################################################
# pcollections/test/_set.py
# Declaration of tests for the pset and tset types.
# By Noah C. Benson

from random import randint
from unittest import TestCase

from .._set import (pset, tset)

class TestPSet(TestCase):
    """Tests for the `pset` and `tset` classes.

    This both runs a number of simple tests of the `pset` API and a series of
    randomized tests in which `pset` must match the behavior of Python's
    native `set` type.
    """
    def test_api(self):
        """Tests/demonstrates the basic pset API."""
        # An empty pset can be created with `pset()`.
        e = pset()
        self.assertEqual(len(e), 0)
        self.assertIsInstance(e, pset)
        # This is identical to the pset.empty object.
        self.assertIs(e, pset.empty)
        # A pset can also be created by passing an iterable to the type.
        p1 = pset(range(10))
        self.assertEqual(len(p1), 10)
        # psets can be equal to each other and to other sets (but, like with
        # set, not to other sequences like tuples or strings).
        l1 = set(range(10))
        self.assertEqual(l1, p1)
        self.assertEqual(p1, l1)
        self.assertEqual(p1, pset(range(10)))
        self.assertNotEqual(p1, tuple(p1))
        self.assertNotEqual(pset.empty, '')
        # psets can perform subset comparisonm including the isdisjoint method.
        self.assertLessEqual(p1, p1)
        self.assertLess(p1, set(range(15)))
        self.assertGreater(p1, pset(range(3,7)))
        self.assertGreaterEqual(p1, p1)
        self.assertTrue(p1.isdisjoint(pset(range(20,30))))
        self.assertFalse(p1.isdisjoint(set(range(-5,5))))
        # Clearing a pset always yields the empty pset.
        self.assertIs(p1.clear(), pset.empty)
        # Copying a pset always just returns the pset (it is immutable).
        self.assertIs(p1.copy(), p1)
        # psets are hashable.
        self.assertIsInstance(hash(p1), int)
        # psets contain values.
        self.assertIn(8, p1)
        self.assertNotIn(11, p1)
        # They can be iterated.
        self.assertEqual(p1, set(iter(p1)))
        # They have lengths.
        self.assertEqual(len(p1), 10)
        # Mutating elements is done by creating a copy with or without an item
        # included using the add or discard methods. This leaves the original
        # object unchanged.
        p2 = p1.add(10)
        self.assertNotIn(10, p1)
        self.assertIn(10, p2)
        p2 = p1.discard(9)
        self.assertIn(9, p1)
        self.assertNotIn(9, p2)
        # The remove method can also be used; it throws an error when an item is
        # not already in the set while discard simply returns the set.
        p2 = p1.remove(9)
        self.assertIn(9, p1)
        self.assertNotIn(9, p2)
        with self.assertRaises(KeyError):
            p1.remove(10)
        p2 = p1.discard(20)
        self.assertIs(p1, p2)
        # The pop method can return an arbitrary element as well as a copy of
        # the pset with that element removed.
        (el, p2) = p1.pop()
        self.assertIn(el, p1)
        self.assertNotIn(el, p2)
        self.assertEqual(p2.add(el), p1)
        # There are also batch methods for adding and removing elements.
        p2 = p1.discardall([4,5,6,7,8,9])
        self.assertEqual(set([0,1,2,3]), p2)
        self.assertEqual(p2.addall(range(4,10)), p1)
        self.assertEqual(p2, p1.removeall(range(4,10)))
        # Basic operators exist for psets as well.
        p1 = pset((1,2,3))
        p2 = pset((2,3,4))
        p3 = pset((5,6,7))
        self.assertEqual(p1 | p2, pset(range(1,5)))
        self.assertEqual(p1 | p3, set((1,2,3,5,6,7)))
        self.assertEqual(p3 - p2, p3)
        self.assertEqual(p2 - p1, set([4]))
        self.assertEqual(p1 & p2, set([2,3]))
        self.assertIs(p1 & p3, pset.empty)
        self.assertTrue(p1.isdisjoint(p3))
        self.assertFalse(p1.isdisjoint((1,2)))
        self.assertTrue(p2.issuperset((3,)))
        self.assertTrue(p2.issuperset((3,4)))
        self.assertTrue(p2.issuperset((2,3,4)))
        self.assertFalse(p2.issuperset((3,4,5)))
        self.assertTrue(p2.issubset((2,3,4)))
        self.assertTrue(p2.issubset((2,3,4,5)))
        self.assertFalse(p2.issubset((2,3)))
        self.assertEqual(p1.difference(p2), set((1,)))
        self.assertEqual(p1.difference(p3), p1)
        self.assertEqual(p2.difference(p1), set((4,)))
        self.assertEqual(p3.difference(p1), p3)
        self.assertEqual(p1.symmetric_difference(p2), set((1,4)))
        self.assertEqual(p1.symmetric_difference(p3), set((1,2,3,5,6,7)))
        self.assertEqual(p2.symmetric_difference(p1), set((1,4)))
        self.assertEqual(p3.symmetric_difference(p1), set((1,2,3,5,6,7)))
        self.assertEqual(p1.intersection(p2), set((2,3)))
        self.assertEqual(p1.intersection(p3), set())
        self.assertEqual(p2.intersection(p1), set((2,3)))
        self.assertEqual(p3.intersection(p1), set())
        self.assertEqual(p1.union(p2), set((1,2,3,4)))
        self.assertEqual(p1.union(p3), set((1,2,3,5,6,7)))
        self.assertEqual(p2.union(p1), set((1,2,3,4)))
        self.assertEqual(p3.union(p1), set((1,2,3,5,6,7)))
    def test_immutable(self):
        """Ensures that `pset` throws the right errors when one mutates it."""
        l = pset(range(10))
        # Cannot set-attr.
        with self.assertRaises(TypeError):
            l._top = -10
    def test_random(self):
        "Performs a randomized test on the pset type."
        nops = 400
        valmax = 20
        p = pset()
        t = tset()
        l = set()
        for opnum in range(nops):
            op = randint(0, 3)
            el = randint(0, valmax)
            if op < 2:
                #print(f"add {el} at {len(l)}")
                p = p.add(el)
                t.add(el)
                l.add(el)
            elif op < 3:
                p = p.discard(el)
                t.discard(el)
                l.discard(el)
            else:
                if el in l:
                    p = p.remove(el)
                    t.remove(el)
                    l.remove(el)
                else:
                    with self.assertRaises(KeyError):
                        p.remove(el)
                    with self.assertRaises(KeyError):
                        t.remove(el)
                    with self.assertRaises(KeyError):
                        l.remove(el)
            #print('   -', repr(p), ' [', p._start, ']')
            #print('   -', repr(t), ' [', t._start, ']')
            #print('   -', repr(l))
            self.assertEqual(p, t)
            self.assertEqual(p, l)
            self.assertEqual(l, t)
            if randint(0, 10) == 0:
                tmp = t.persistent()
                self.assertEqual(tmp, p)
                self.assertEqual(tmp, t)
                for k in l:
                    self.assertIn(k, p)
                    self.assertIn(k, t)
                    self.assertIn(k, tmp)



# -*- coding: utf-8 -*-
################################################################################
# pcollections/abc/__init__.py
# Initialization file for the pcollections library abstract base class core.
# By Noah C. Benson

"""Abstract base classes for persistent and transient collections for Python.
"""

from ._core import (Persistent,         Transient)
from ._seq  import (PersistentSequence, TransientSequence)
from ._set  import (PersistentSet,      TransientSet)
from ._map  import (PersistentMapping,  TransientMapping)
